@@ -1,5 +1,9 @@
 #! /usr/bin/env python3
 
+"""
+This is the main part of xyz_py
+"""
+
 import numpy as np
 import numpy.linalg as la
 from ase import neighborlist, Atoms
@@ -7,8 +11,9 @@ from ase.geometry.analysis import Analysis
 import copy
 import re
 import scipy.optimize as spo
-import itertools
 import sys
+from collections import defaultdict
+
 
 atom_lab_num = {"H": 1, "He": 2, "Li": 3, "Be": 4, "B": 5, "C": 6, "N": 7,
                 "O": 8, "F": 9, "Ne": 10, "Na": 11, "Mg": 12, "Al": 13,
@@ -87,7 +92,7 @@ def load_xyz(f_name: str, atomic_numbers: bool = False,
     return _labels, _coords
 
 
-def save_xyz(f_name: str, labels: list, coords: list,
+def save_xyz(f_name: str, labels: list, coords: np.ndarray,
              with_numbers: bool = False, verbose: bool = True,
              mask: list = [], atomic_numbers: bool = False):
     """
@@ -99,7 +104,7 @@ def save_xyz(f_name: str, labels: list, coords: list,
         File name
     labels : list
         atomic labels
-    coords : list
+    coords : np.ndarray
         list of 3 element lists containing xyz coordinates of each atom
     with_numbers : bool, default False
         If True, add/overwrite numbers to labels before printing
@@ -124,16 +129,12 @@ def save_xyz(f_name: str, labels: list, coords: list,
     else:
         _labels = labels
 
-    if len(mask) != 0:
-        n_atoms = int(sum(mask))
-        _mask = mask
-    else:
-        n_atoms = len(_labels)
-        _mask = [1]*n_atoms
+    # Set up masks
+    if mask:
+        coords = np.delete(coords, mask, axis=0)
+        _labels = np.delete(_labels, mask, axis=0).tolist()
 
-    # Apply mask
-    _labels = _mask_list(_labels, _mask)
-    _coords = _mask_list(coords, _mask)
+    n_atoms = len(_labels)
 
     if atomic_numbers:
         _labels = remove_numbers(_labels)
@@ -144,7 +145,7 @@ def save_xyz(f_name: str, labels: list, coords: list,
 
     with open(f_name, 'w') as f:
         f.write("{:d}\n\n".format(n_atoms))
-        for it, (ident, trio) in enumerate(zip(_identifier, _coords)):
+        for it, (ident, trio) in enumerate(zip(_identifier, coords)):
             f.write("{:5} {:15.7f} {:15.7f} {:15.7f} \n".format(ident, *trio))
 
     if verbose:
@@ -432,18 +433,18 @@ def combine_xyz(labels_1: list, labels_2: list,
     labels_1 : list
         Atomic labels
     coords_1 : list
-        list of lists of xyz coordinates of each atom
+        xyz coordinates as (n_atoms, 3) array
     labels_2 : list
         Atomic labels
     coords_2 : list
-        list of lists of xyz coordinates of each atom
+        xyz coordinates as (n_atoms, 3) array
 
     Returns
     -------
     list
         Combined atomic labels
-    list
-        Combined list of lists of xyz coordinates of each atom
+    np.ndarray
+        Combined xyz coordinates as (n_atoms, 3) array
     """
 
     # Concatenate labels lists
@@ -455,16 +456,17 @@ def combine_xyz(labels_1: list, labels_2: list,
     return labels, coords
 
 
-def get_neighborlist(labels: list, coords: list, adjust_cutoff: dict = {}):
+def get_neighborlist(labels: list, coords: np.ndarray,
+                     adjust_cutoff: dict = {}):
     """
-    Calculate ASE neighbourlist with cutoffs
+    Calculate ASE neighbourlist based on covalent radii
 
     Parameters
     ----------
     labels : list
         Atomic labels
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     adjust_cutoff : dict, optional
         dictionary of atoms (keys) and new cutoffs (values)
 
@@ -500,18 +502,16 @@ def get_neighborlist(labels: list, coords: list, adjust_cutoff: dict = {}):
     return neigh_list
 
 
-def get_adjacency(labels: list, coords: list, adjust_cutoff: dict = {},
-                  save: bool = False, f_name: str = "adjacency.dat"):
+def get_adjacency(labels: list, coords: np.ndarray, adjust_cutoff: dict = {}):
     """
-    Calculate adjacency matrix using ASE built in cutoffs with option to
-    modify them
+    Calculate adjacency matrix using ASE based on covalent radii.
 
     Parameters
     ----------
     labels : list
         Atomic labels
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     adjust_cutoff : dict, optional
         dictionary of atoms (keys) and new cutoffs (values)
     save : bool, default False
@@ -536,28 +536,22 @@ def get_adjacency(labels: list, coords: list, adjust_cutoff: dict = {},
     # Create adjacency matrix
     adjacency = neigh_list.get_connectivity_matrix(sparse=False)
 
-    # Save adjacency matrix to file
-    if save:
-        np.savetxt(f_name, adjacency, fmt="%i")
-        print('Adjacency matrix saved to {}'.format(f_name))
-
     return adjacency
 
 
-def get_bonds(labels: list, coords: list, neigh_list=None,
-              f_name: str = 'bonds.dat', save: bool = False,
+def get_bonds(labels: list, coords: np.ndarray, neigh_list=None,
               verbose: bool = True, style: bool = 'indices'):
     """
-    Calculate and save list of atoms between which there is a bond.
-    Using ASE. Only unique angles are saved.
+    Calculate list of atoms between which there is a bond.
+    Using ASE. Only unique angles are retained.
     e.g. 0-1 and not 1-0
 
     Parameters
     ----------
     labels : list
         Atomic labels
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     neigh_list : ASE neighbourlist object, optional
         neighbourlist of system
     f_name : str, 'bonds.dat'
@@ -610,19 +604,11 @@ def get_bonds(labels: list, coords: list, neigh_list=None,
     # Count bonds
     n_bonds = len(bonds)
 
-    # Set save format and convert to atomic labels if requested
-    if style == 'indices':
-        save_fmt = "%i"
-    elif style == 'labels':
+    # Set format and convert to atomic labels if requested
+    if style == 'labels':
         bonds = [[labels[atom1], labels[atom2]] for atom1, atom2 in bonds]
-        save_fmt = "%s"
     else:
-        sys.exit("Unknown save style specified")
-
-    # Save bond list to file
-    if save:
-        np.savetxt(f_name, bonds, fmt=save_fmt)
-        print('Bond list saved to {}'.format(f_name))
+        sys.exit("Unknown style specified")
 
     # Print number of bonds to screen
     if verbose:
@@ -631,20 +617,19 @@ def get_bonds(labels: list, coords: list, neigh_list=None,
     return bonds
 
 
-def get_angles(labels: list, coords: list, neigh_list=None,
-               f_name: str = 'angles.dat', save: bool = False,
+def get_angles(labels: list, coords: np.ndarray, neigh_list=None,
                verbose: bool = True, style: bool = 'indices'):
     """
-    Calculate and save list of atoms between which there is a bond angle.
-    Using ASE. Only unique angles are saved.
+    Calculate list of atoms between which there is a bond angle.
+    Using ASE. Only unique angles are retained.
     e.g. 0-1-2 but not 2-1-0
 
     Parameters
     ----------
     labels : list
         Atomic labels
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     neigh_list : ASE neighbourlist object, optional
         neighbourlist of system
     f_name : str, default 'angles.dat'
@@ -697,20 +682,12 @@ def get_angles(labels: list, coords: list, neigh_list=None,
     # Count angles
     n_angles = len(angles)
 
-    # Set save format and convert to atomic labels if requested
-    if style == 'indices':
-        save_fmt = "%i"
-    elif style == 'labels':
+    # Set format and convert to atomic labels if requested
+    if style == 'labels':
         angles = [[labels[atom1], labels[atom2], labels[atom3]]
                   for atom1, atom2, atom3 in angles]
-        save_fmt = "%s"
     else:
-        sys.exit("Unknown save style specified")
-
-    # Save angle list to file
-    if save:
-        np.savetxt(f_name, angles, fmt=save_fmt)
-        print('Angle list saved to {}'.format(f_name))
+        sys.exit("Unknown style specified")
 
     # Print number of angles to screen
     if verbose:
@@ -719,20 +696,19 @@ def get_angles(labels: list, coords: list, neigh_list=None,
     return angles
 
 
-def get_dihedrals(labels: list, coords: list, neigh_list=None,
-                  f_name: str = 'dihedrals.dat', save: bool = False,
+def get_dihedrals(labels: list, coords: np.ndarray, neigh_list=None,
                   verbose: bool = True, style: bool = 'indices'):
     """
-    Calculate and save list of atoms between which there is a dihedral.
-    Using ASE. Only unique angles are saved.
+    Calculate and list of atoms between which there is a dihedral.
+    Using ASE. Only unique angles are retained.
     e.g. 0-1-2-3 but not 3-2-1-0
 
     Parameters
     ----------
     labels : list
         Atomic labels
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     neigh_list : ASE neighbourlist object, optional
         neighbourlist of system
     f_name : str, default 'dihedrals.dat'
@@ -785,10 +761,8 @@ def get_dihedrals(labels: list, coords: list, neigh_list=None,
     # Count dihedrals
     n_dihedrals = len(dihedrals)
 
-    # Set save format and convert to atomic labels if requested
-    if style == 'indices':
-        save_fmt = "%i"
-    elif style == 'labels':
+    # Set format and convert to atomic labels if requested
+    if style == 'labels':
         dihedrals = [
                      [
                       labels[atom1],
@@ -798,14 +772,8 @@ def get_dihedrals(labels: list, coords: list, neigh_list=None,
                      ]
                      for atom1, atom2, atom3, atom4 in dihedrals
                     ]
-        save_fmt = "%s"
     else:
-        sys.exit("Unknown save style specified")
-
-    # Save dihedral list to file
-    if save:
-        np.savetxt(f_name, dihedrals, fmt=save_fmt)
-        print('Dihedral list saved to {}'.format(f_name))
+        sys.exit("Unknown style specified")
 
     # Print number of dihedrals to screen
     if verbose:
@@ -863,19 +831,19 @@ def num_to_lab(numbers: list, numbered: bool = True):
     return labels_wn
 
 
-def reflect_coords(coords: list):
+def reflect_coords(coords: np.ndarray):
     """
     Reflect coordinates through xy plane
 
     Parameters
     ----------
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
 
     Returns
     -------
-    list
-        list of lists of xyz coordinates of each atom
+    np.ndarray
+        xyz coordinates as (n_atoms, 3) array
 
     """
 
@@ -904,23 +872,97 @@ def reflect_coords(coords: list):
     return coords
 
 
-def remove_broken(labels: list, coords: list, formulae: list,
+def find_entities(labels: list, coords: np.ndarray, adjust_cutoff: dict = {}):
+    """
+    Finds formulae of entities given in labels and coords using adjacency
+    matrix
+
+    Parameters
+    ----------
+    labels : list
+        atomic labels
+    coords : np.ndarray
+        xyz coordinates of each atom as (n_atoms, 3) array
+    adjust_cutoff : dict, optional
+        dictionary of atoms (keys) and new cutoffs (values) used in generating
+        adjacency matrix
+
+    Returns
+    -------
+    dict
+        keys = molecular formula,
+        vals = list of lists, where each list contains the indices of a single
+                occurrence of the `key`, and the indices match the order given
+                in `labels` and `coords`
+    """
+
+    # Remove label numbers if present
+    _labels = remove_numbers(labels)
+
+    # Generate adjacency matrix using ASE
+    adjacency = get_adjacency(_labels, coords, adjust_cutoff=adjust_cutoff)
+
+    # Count number of atoms
+    n_atoms = len(labels)
+
+    # Set current fragment as start atom
+    curr_frag = {0}
+
+    # List of unvisited atoms
+    unvisited = set(np.arange(n_atoms))
+
+    # Dictionary of molecular_formula:[[indices_mol1], [indices_mol2]] pairs
+    mol_indices = defaultdict(list)
+
+    # Loop over adjacency matrix and trace out bonding network
+    # Make a first pass, recording in a list the atoms which are bonded to the
+    # first atom.
+    # Then make another pass, and record in another list all the atoms bonded
+    # to those in the previous list
+    # and again, and again etc.
+    while unvisited:
+        # Keep copy of current fragment indices to check against for changes
+        prev_frag = copy.copy(curr_frag)
+        for index in prev_frag:
+            # Find bonded atoms and add to current fragment
+            indices = list(np.nonzero(adjacency[:, index])[0])
+            curr_frag.update(indices)
+
+        # If no changes in fragment last pass, then a complete structure must
+        # have been found
+        if prev_frag == curr_frag:
+
+            # Generate molecular formula of current fragment
+            curr_labels = [_labels[it] for it in curr_frag]
+            curr_formula = count_elements(curr_labels)
+
+            mol_indices[formdict_to_formstr(curr_formula)].append(
+                list(curr_frag)
+            )
+
+            # Remove visited atoms
+            unvisited = unvisited.difference(curr_frag)
+
+            # Reset lists of labels and indices ready for next cycle
+            curr_frag = {min(unvisited)} if unvisited else curr_frag
+
+    return dict(mol_indices)
+
+
+def remove_broken(labels: list, coords: np.ndarray, formulae: list,
                   adjust_cutoff: dict = {}, verbose: bool = False,
                   save: bool = False, frag_f_name: str = "fragments.xyz",
                   clean_f_name: str = "clean.xyz",
                   mask: list = [], skip: list = []):
     """
-    Removes structures from coords which do not match formula in formulae
-    Uses ASE to generate adjacency matrix and draw out bonding in coords
-    allowing formulae of fragments to be found.
-
+    Deprecated, use find_entities instead.
 
     Parameters
     ----------
     labels : list
         list of atomic labels
-    coords : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     formulae : list
         list of chemical formulae stored as dictionaries with atomic
         symbol (key) count (val) pairs
@@ -940,256 +982,40 @@ def remove_broken(labels: list, coords: list, formulae: list,
     skip : list, optional
         List of atomic indices which shall not be visited when tracing
         bonding network
+
     Returns
     -------
     list
         atomic labels for full molecules
-    list
-        list of lists of xyz coordinates of each atom for full molecules
+    np.ndarray
+        xyz coordinates as (n_atoms, 3) array for full molecules
     list
         atomic labels for fragments
-    list
-        list of lists of xyz coordinates of each atom for fragments
+    np.ndarray
+        xyz coordinates as (n_atoms, 3) array for fragments
     dict
         keys = molecular formula,
         vals = list of lists of atomic indices
                 for each atom
     """
 
-    # Remove label numbers if present
-    _labels = remove_numbers(labels)
-
-    n_atoms = len(labels)
-
-    # Set up array of atoms to ignore
-    # This becomes the array of atomic indices which have been
-    # visited
-    if skip:
-        visited = set(skip)
-    else:
-        visited = set()
-
-    # Set up mask
-    if mask:
-        _mask = mask
-    else:
-        _mask = [1]*n_atoms
-
-    # Apply mask
-    _labels = _mask_list(_labels, _mask)
-    _coords = _mask_list(coords, _mask)
-
-    # Generate adjacency matrix using ASE
-    adjacency = get_adjacency(_labels, _coords, adjust_cutoff=adjust_cutoff)
-
-    # Count number of atoms
-    n_atoms = len(labels)
-
-    # Mask for fragment atom indices
-    # 1 = Fragment atom
-    # 0 = Not a fragment atom
-    frag_mask = np.zeros(n_atoms)
-
-    # Set starting position as first unvisited/unskipped atom
-    unvisited = [ind for ind, vis
-                 in zip(np.arange(n_atoms), sorted(visited)) if ind != vis]
-    if unvisited:
-        start = min(unvisited)
-    else:
-        start = 0
-    # Set current fragment as start atom
-    curr_frag = {start}
-
-    # Dictionary of molecular_formula:count pairs
-    num_molecules = {formdict_to_formstr(form): 0 for form in formulae}
-    # Dictionary of molecular_formula:[[indices_mol1], [indices_mol2]] pairs
-    mol_indices = {formdict_to_formstr(form): [] for form in formulae}
-
-    # Loop over adjacency matrix and trace out bonding network
-    # Make a first pass, recording in a list the atoms which are bonded to the
-    # first atom.
-    # Then make another pass, and record in another list all the atoms bonded
-    # to those in the previous list
-    # and again, and again etc.
-    while len(visited) != n_atoms:
-        # Keep copy of current fragment indices to check against for changes
-        prev_frag = copy.copy(curr_frag)
-        for index in prev_frag:
-            # Find bonded atoms and add to current fragment
-            indices = list(np.nonzero(adjacency[:, index])[0])
-            curr_frag.update(indices)
-
-        # If no changes in fragment last pass, then a complete structure must
-        # have been found
-        if prev_frag == curr_frag:
-
-            # Keep list of all atom indices which have been visited
-            visited.update(curr_frag)
-
-            # Generate molecular formula of current fragment
-            curr_labels = [_labels[it] for it in curr_frag]
-            curr_formula = count_elements(curr_labels)
-
-            # Check if formula matches any of user provided formulae
-            if any(form == curr_formula for form in formulae):
-                num_molecules[formdict_to_formstr(curr_formula)] += 1
-                # Reorder indices to have same label order as formula
-                # e.g. H2O --> H, H, O
-                curr_frag = list(curr_frag)
-                ord_frag = [
-                            curr_frag[o]
-                            for o in np.argsort(lab_to_num(curr_labels))
-                           ]
-                mol_indices[formdict_to_formstr(curr_formula)].append(ord_frag)
-            # If there are no matches, then mark the fragment as a fragment
-            else:
-                for it in curr_frag:
-                    frag_mask[it] = 1
-
-            # Move starting point to next lowest index which has
-            # not been visited
-            unvisited = [
-                         ind
-                         for ind in np.arange(n_atoms) if ind not in visited
-                        ]
-            if not unvisited:
-                break
-
-            # Reset lists of labels and indices ready for next cycle
-            curr_frag = {min(unvisited)}
-
-    # Print information on how many of each formula have been found
-    if verbose:
-        for mol in num_molecules.keys():
-            print("There is/are {:10d} molecule(s)/ion(s) of {}".format(
-                num_molecules[mol], mol)
-            )
-
-    # Get indices of all molecules
-    # Exclude molecules that were not found in xyz file
-    # i.e. inds_clean is a list of lists
-    # [
-    #  [indices_of_all_instances_of_molecule1],
-    #  [indices_of_all_instances_of_molecule2]
-    # ]
-    inds_clean = [
-                  [item for sublist in molecule for item in sublist]
-                  for molecule in mol_indices.values() if len(molecule) > 0
-                 ]
-    # Get all molecular formulae in a list
-    form_clean = [
-                  formula
-                  for formula in mol_indices.keys()
-                  if len(mol_indices[formula]) > 0
-                 ]
-
-    # Reorder inds_clean so that metal containing molecules are at the start
-    metals_top_order = np.argsort(
-        [not int(contains_metal(formula)) for formula in form_clean]
-        )
-    inds_clean = _reorder_list(inds_clean, metals_top_order)
-
-    # Concatenate inds_clean into a single big list
-    # [
-    #  indices_of_all_instances_of_molecule1,
-    #  indices_of_all_instances_of_molecule2
-    # ]
-    inds_clean = list(itertools.chain(*inds_clean))
-
-    # Create output array of labels and coordinates where molecules
-    # are grouped together, and
-    # metal containing molecules are at the top using inds_clean to
-    # set that order
-    labels_clean = _reorder_list(_labels, inds_clean)
-    coords_clean = _reorder_list(_coords, inds_clean)
-
-    # Mask off molecules to get output array of fragments
-    # order does not matter here
-    labels_fragments = _mask_list(_labels, frag_mask)
-    coords_fragments = _mask_list(_coords, frag_mask)
-
-    # Save clean and fragmented coordinates to file
-    if save:
-        # Write new xyz file without broken molecules
-        # Save xyz for complete molecules
-        save_xyz(clean_f_name, labels_clean, coords_clean)
-        # Save xyz for fragments
-        save_xyz(frag_f_name, labels_fragments, coords_fragments)
-
-    return [
-        labels_clean, coords_clean,
-        labels_fragments, coords_fragments, mol_indices
-        ]
+    print("xyz_py: remove_broken is deprecated, use find_entities instead")
+    exit()
+    return
 
 
-def _mask_list(_list: list, mask: list, opp_mask: bool = False):
-    """
-    Truncates list to only include elements specified by 1 in mask
-
-    Parameters
-    ----------
-    _list : list
-        list to mask
-    mask : list
-        list of 0 (exclude) and 1 (include) for each element
-
-    opp_mask : bool, optional
-        Apply mask in opposite fashion:
-            0 (include) and 1 (exclude)
-
-    Returns
-    -------
-    float
-        Masked list
-    """
-
-    # Flip mask other way around if requested
-    if opp_mask:
-        _mask = [not ma for ma in mask]
-    else:
-        _mask = mask
-
-    # Apply mask to list
-    _list_masked = [_l for it, _l in enumerate(_list) if _mask[it]]
-
-    return _list_masked
-
-
-def _reorder_list(_list: list, order: list):
-    """
-    Reorders list using indices in order
-
-    Parameters
-    ----------
-    _list : list
-        list to reorder
-    order : list
-        New order for list
-
-    Returns
-    -------
-    list
-        Reordered list
-    """
-
-    # Apply mask to list
-    _list_reordered = [_list[it] for it in order]
-
-    return _list_reordered
-
-
-def _calculate_rmsd(coords_1: list, coords_2: list):
+def _calculate_rmsd(coords_1: np.ndarray, coords_2: np.ndarray):
     """
     Calculates RMSD between two structures
     RMSD = sqrt(mean(deviations**2))
     Where deviations are defined as norm([x1,y1,z1]-[x2,y2,z2])
 
     Parameters
-----------
-    coords_1 : list
-        list of lists of xyz coordinates of each atom
-    coords_2 : list
-        list of lists of xyz coordinates of each atom
+    ----------
+    coords_1 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
+    coords_2 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
 
     Returns
     -------
@@ -1215,8 +1041,9 @@ def _calculate_rmsd(coords_1: list, coords_2: list):
     return rmsd
 
 
-def calculate_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
-                   mask_2: list = [], order_1: list = [], order_2: list = []):
+def calculate_rmsd(coords_1: np.ndarray, coords_2: np.ndarray,
+                   mask_1: list = [], mask_2: list = [], order_1: list = [],
+                   order_2: list = []):
     """
     Calculates RMSD between two structures
     RMSD = sqrt(mean(deviations**2))
@@ -1228,10 +1055,10 @@ def calculate_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
 
     Parameters
     ----------
-    coords_1 : list
-        list of lists of xyz coordinates of each atom
-    coords_2 : list
-        list of lists of xyz coordinates of each atom
+    coords_1 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
+    coords_2 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
 
     mask_1 : list
         list of 0 (exclude) and 1 (include) for each element in coords_1
@@ -1260,24 +1087,16 @@ def calculate_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
         _order_2 = range(len(coords_2))
 
     # Apply new order
-    _coords_1 = _reorder_list(coords_1, _order_1)
-    _coords_2 = _reorder_list(coords_2, _order_2)
+    _coords_1 = coords_1[_order_1]
+    _coords_2 = coords_2[_order_2]
 
     # Set up masks
     if mask_1:
-        _mask_1 = mask_1
-    else:
-        _mask_1 = [1]*len(coords_1)
+        _coords_1 = np.delete(_coords_1, mask_1, axis=0)
 
     # Set up masks
     if mask_2:
-        _mask_2 = mask_2
-    else:
-        _mask_2 = [1]*len(coords_2)
-
-    # Apply masks
-    _coords_1 = _mask_list(_coords_1, _mask_1)
-    _coords_2 = _mask_list(_coords_2, _mask_2)
+        _coords_2 = np.delete(_coords_2, mask_2, axis=0)
 
     # Calculate rmsd
     rmsd = _calculate_rmsd(_coords_1, _coords_2)
@@ -1285,15 +1104,15 @@ def calculate_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
     return rmsd
 
 
-def rotate_coords(coords: list, alpha: float, beta: float, gamma: float):
+def rotate_coords(coords: np.ndarray, alpha: float, beta: float, gamma: float):
     """
     Rotates coordinates by alpha, beta, gamma using the zyz convention
     https://easyspin.org/easyspin/documentation/eulerangles.html
 
     Parameters
     ----------
-    coords_1 : list
-        list of lists of xyz coordinates of each atom
+    coords : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     alpha : float
         alpha angle in radians
     beta : float
@@ -1303,8 +1122,8 @@ def rotate_coords(coords: list, alpha: float, beta: float, gamma: float):
 
     Returns
     -------
-    list
-        list of lists of xyz coordinates of each atom after rotation
+    np.ndarray
+        xyz coordinates as (n_atoms, 3) array after rotation
         in same order as input coordinates
     """
 
@@ -1322,7 +1141,7 @@ def rotate_coords(coords: list, alpha: float, beta: float, gamma: float):
     R[2, 2] = np.cos(beta)
 
     # Create (n,3) matrix from coords list
-    _coords = np.asarray(coords).T
+    _coords = coords.T
 
     # Apply rotation matrix
     rot_coords = R @ _coords
@@ -1330,27 +1149,25 @@ def rotate_coords(coords: list, alpha: float, beta: float, gamma: float):
     # Convert back to (3,n) matrix
     rot_coords = rot_coords.T
 
-    # Convert back to list
-    rot_coords = list(rot_coords)
-
     return rot_coords
 
 
-def minimise_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
-                  mask_2: list = [], order_1: list = [], order_2: list = []):
+def minimise_rmsd(coords_1: np.ndarray, coords_2: np.ndarray,
+                  mask_1: list = [], mask_2: list = [], order_1: list = [],
+                  order_2: list = []):
     """
     Minimising the RMSD between two structures
     If coords_1 and coords_2 are not the same length, then a mask array can be
     provided for either/both and is applied prior to the calculation
     coords_1 and coords_2 can also be reordered if new orders are specified
-        - note this occurs BEFORE masking
+    **note reordering occurs before masking**
 
     Parameters
     ----------
-    coords_1 : list
-        list of lists of xyz coordinates of each atom
-    coords_2 : list
-        list of lists of xyz coordinates of each atom
+    coords_1 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
+    coords_2 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array
     mask_1 : list
         list of 0 (exclude) and 1 (include) for each element in coords_1
     mask_2 : list
@@ -1378,28 +1195,20 @@ def minimise_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
         _order_2 = range(len(coords_2))
 
     # Apply new order
-    _coords_1 = _reorder_list(coords_1, _order_1)
-    _coords_2 = _reorder_list(coords_2, _order_2)
+    _coords_1 = coords_1[_order_1]
+    _coords_2 = coords_2[_order_2]
 
     # Set up masks
     if mask_1:
-        _mask_1 = mask_1
-    else:
-        _mask_1 = [1]*len(coords_1)
+        _coords_1 = np.delete(_coords_1, mask_1, axis=0)
 
     # Set up masks
     if mask_2:
-        _mask_2 = mask_2
-    else:
-        _mask_2 = [1]*len(coords_2)
-
-    # Apply masks
-    _coords_1 = _mask_list(_coords_1, _mask_1)
-    _coords_2 = _mask_list(_coords_2, _mask_2)
+        _coords_2 = np.delete(_coords_2, mask_2, axis=0)
 
     # Fit alpha, beta, and gamma to minimise rmsd
     result = spo.least_squares(lambda angs: _rotate_and_rmsd(
-            angs, _coords_1, _coords_2), x0=(1., 1., 1.))
+            angs, _coords_1, _coords_2), x0=(1., 1., 1.), jac='3-point')
 
     # Get optimum angles
     [alpha, beta, gamma] = result.x
@@ -1408,7 +1217,7 @@ def minimise_rmsd(coords_1: list, coords_2: list, mask_1: list = [],
     return rmsd, alpha, beta, gamma
 
 
-def _rotate_and_rmsd(angs: list, coords_1: list, coords_2: list):
+def _rotate_and_rmsd(angs: list, coords_1: np.ndarray, coords_2: np.ndarray):
     """
     Rotates coords_1 by alpha, beta, gamma using the zyz convention
     https://easyspin.org/easyspin/documentation/eulerangles.html
@@ -1416,17 +1225,17 @@ def _rotate_and_rmsd(angs: list, coords_1: list, coords_2: list):
 
     Parameters
     ----------
-    coords_1 : list
-        list of lists of xyz coordinates of each atom of first system
-    coords_2 : list
-        list of lists of xyz coordinates of each atom of second system
+    coords_1 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array of first system
+    coords_2 : np.ndarray
+        xyz coordinates as (n_atoms, 3) array of second system
     angs : list
         alpha, beta, gamma in radians
 
     Returns
     -------
-    list
-        list of lists of xyz coordinates of each atom after rotation
+    np.ndarray
+        xyz coordinates as (n_atoms, 3) array after rotation
         in same order as input coordinates
     """
 
